import csv
import pathlib
import random

from . import reference as seedweed


def load():
    data_file = pathlib.Path(__file__).parent / "test-vectors.csv"
    reader = csv.DictReader(open(data_file))
    data = []
    for row in reader:
        credential_id = bytes.fromhex(row["credential_id"])
        nonce, extstate, mac = seedweed.nonce_extstate_mac_from_credential_id(
            credential_id
        )
        data.append(
            {
                "seed": bytes.fromhex(row["seed"]),
                "rp_id": row["rp_id"],
                "nonce": nonce,
                "mac": mac,
                "credential_id": credential_id,
                "secret_scalar": row["sec_scalar"],
                "ext_state": extstate,
            }
        )

    return data


def random_bytes(rng, length):
    return rng.getrandbits(8 * length).to_bytes(length, "little")


P256 = seedweed.P256
H = seedweed.H


class Parameters:
    def __init__(self):
        # independent DRNG
        rng = random.Random(2020)

        self.seeds = [
            random_bytes(rng, 32),
            bytes.fromhex(
                "dd2ca3b88f9491c042fcc04c5e732e9f6fd9c0eb6f3b99cddd4ae96d661ada2c"
            ),
            bytes([0] * 32),
            (P256.order + 1).to_bytes(32, "big"),
            bytes([1] * 32),
            bytes([0xFF] * 32),
            H(b"the dicekey method"),
        ]

        self.rp_ids = [
            "solokeys.com",
            "dicekeys.com",
            "fidoalliance.org",
            "example.com",
        ]

        self.nonces = [random_bytes(rng, 32) for _ in range(4)] + [
            bytes.fromhex(
                "8b911917c0b74f77e6fb819d6a8034a94f0fca487cacb41e89235c5c5220947b"
            )
        ]
        self.extra_states = [
            b"",
            b"sunny side up",
            random_bytes(rng, 256),
        ]


def generate(parameters=Parameters()):

    print(
        "".join(
            (
                "seed,rp_id,nonce,mac,credential_id,sec_scalar,pub_key,",
                "example_signature_for_seedweed,iterations",
            )
        )
    )

    import itertools

    max_iterations = 0
    for extra_state, seed, nonce, rp_id in itertools.product(
        parameters.extra_states,
        parameters.seeds,
        parameters.nonces,
        parameters.rp_ids,
    ):
        rp_id_hash = H(rp_id.encode())

        credential_id = seedweed.credential_id_from_seed_nonce_rpidhash(
            seed,
            nonce,
            rp_id_hash,
            extra_state,
        )

        assert seedweed.validate_credential_id(seed, credential_id, rp_id_hash)

        mac = credential_id[-32:]

        scalar, point, keypair, iterations = seedweed.keypair_from_seed_mac(seed, mac)
        # TODO: we really need some examples with >1 iterations,
        # ideally even some rare >2 case
        if iterations > max_iterations:
            max_iterations = iterations
        assert 1 <= scalar < P256.order

        # X big-endian 32B || Y big-endian 32B
        pub_key_uncompressed = keypair.verifying_key._raw_encode()

        signature = keypair.sign_deterministic(b"seedweed")

        nonce, ext_state, mac = seedweed.nonce_extstate_mac_from_credential_id(
            credential_id
        )
        # reformatted_credential_id = ":".join(
        #     ["1", nonce.hex(), mac.hex(), ext_state.hex(), mac.hex()]
        # )

        print(
            "".join(
                (
                    f"{seed.hex()},{rp_id},{nonce.hex()},{mac.hex()},{credential_id.hex()},",
                    f"{scalar},{pub_key_uncompressed.hex()},{signature.hex()},{iterations}",
                )
            )
        )

    # have test cases to cover the 1 in 4 billion case
    assert 1 < max_iterations  # <= 2



from . import reference as seedweed

H = seedweed.H


def verify_make_credential(
    testvector,
    authnr_credential_id,
    authnr_attn_signature,
    attested_data,
):

    seed = testvector["seed"]
    rp_id_hash = H(testvector["rp_id"].encode())
    # check credential ID is well-formed
    seedweed.validate_credential_id(
        testvector["seed"],
        authnr_credential_id,
        rp_id_hash,
    )

    authnr_nonce, _, authnr_mac = seedweed.nonce_extstate_mac_from_credential_id(
        authnr_credential_id
    )

    credential_id = seedweed.credential_id_from_seed_nonce_rpidhash(
        seed, authnr_nonce, rp_id_hash
    )

    # tests mac == authnr_mac
    assert credential_id == authnr_credential_id

    _, _, keypair, _ = seedweed.keypair_from_seed_mac(seed, authnr_mac)

    # check signature, raises `BadSignatureError: Signature verification failed` else
    keypair.verifying_key.verify(authnr_attn_signature, attested_data)


def verify_get_assertion(
    testvector,
    authnr_signature,
    signed_data,
):
    seed = testvector["seed"]
    mac = testvector["mac"]

    _, _, keypair, _ = seedweed.keypair_from_seed_mac(seed, mac)

    # check signature, raises `BadSignatureError: Signature verification failed` else
    keypair.verifying_key.verify(authnr_signature, signed_data)



""" Find setups where the P256 key construction takes more than one interation. """

import hashlib
import hmac
import random
import secrets

import ecdsa
import seedweed

P256 = ecdsa.NIST256p

def random_bytes(length):
    return random.getrandbits(8*length).to_bytes(n, 'little')


def HMAC(key: bytes, data: bytes) -> bytes:
    mac = hmac.new(key, digestmod=hashlib.sha256)
    mac.update(data)
    return mac.digest()

# seed = random_bytes(32)
seed = seedweed.H(b"seedweed")
print(f"seed = {seed.hex()}")

random.seed(2020)
search_seed = secrets.token_bytes(32)
# search_seed = bytes.fromhex("252dbf25c582dfa486f89f692d3c742152e81f35d7240668f2b026a02bd4ae83")
print(f"search seed = {search_seed.hex()}")
rng = random.Random(search_seed)

order = P256.order
found = 0

rp_id = "solokeys.com".encode()
rp_id_hash = seedweed.H(rp_id)
with open(f"{search_seed.hex()}.log", "w") as fh:
    # for i in range(2**32):
    # for i in range(1000000000):
    for i in range(2**64):
        # mac = int(i).to_bytes(32, "big")
        unique_id = rng.getrandbits(256).to_bytes(32, 'big')
        mac = HMAC(seed, rp_id_hash + bytes([1]) + unique_id)
        candidate = int.from_bytes(HMAC(seed, mac), "little")

        if candidate >= order:
            found += 1
            entry = f"i = {i}: HMAC({seed.hex()},HMAC(seed, H(b'example.com') + [1] + {unique_id.hex()})) = {candidate}"
            print(entry)
            fh.write(entry + "\n")
            fh.flush()

            if True:
                fh.close()
                import sys
                sys.exit(0)

        if i % int(1e6) == 0:
            print(f"found {found} after {i}")


# 275edb1c190bd62fc400df5045190cfc647b0956a390685389d7b404a86a3073.log
# HMAC(
#   59ece2f2bcd6512eea7a6bd1694cf1e73552fbf314e219b06f3e2f864f0d99ed,
#   fd7fc7d2b1b49c7cba47dd3e7a5f3cb991bf0b62388e720bd9be272163a6a725,
# ) = 115792089226597647821555558773243515956592285713205569838850865706012713790799




"""Seeded WebAuthn Credentials: reference implementation"""

__version__ = "1.0rc3"

from . import conformance, reference, vectors  # noqa: F401
from .reference import *  # noqa: F401,F403


def load_test_vectors():
    return vectors.load()



import hashlib
import hmac

import ecdsa

P256 = ecdsa.NIST256p

VERSION = bytes([1])


def assert_bytes(value):
    assert isinstance(value, bytes), value


def assert_32_bytes(value):
    assert_bytes(value)
    assert len(value) == 32, value


def assert_credential_id(credential_id: bytes):
    assert_bytes(credential_id)
    assert 65 <= len(credential_id) <= 65 + 256
    assert credential_id[:1] == VERSION


def H(data: bytes) -> bytes:
    assert_bytes(data)

    hasher = hashlib.sha256()
    hasher.update(data)
    return hasher.digest()


def HMAC(key: bytes, data: bytes) -> bytes:
    assert_32_bytes(key)
    assert_bytes(data)

    mac = hmac.new(key, digestmod=hashlib.sha256)
    mac.update(data)
    return mac.digest()


def keypair_from_seed_mac(seed: bytes, mac: bytes) -> ecdsa.SigningKey:
    assert_32_bytes(seed)
    assert_32_bytes(mac)

    P256 = ecdsa.NIST256p

    candidate_bytes = HMAC(seed, mac)
    candidate_scalar = int.from_bytes(candidate_bytes, "little")
    iterations = 1
    while not (0 < candidate_scalar < P256.order):
        candidate_bytes = HMAC(seed, candidate_bytes)
        candidate_scalar = int.from_bytes(candidate_bytes, "little")
        iterations += 1

    secret_scalar = candidate_scalar
    public_point = secret_scalar * P256.generator
    keypair = ecdsa.SigningKey.from_secret_exponent(secret_scalar, P256)

    return secret_scalar, public_point, keypair, iterations


def credential_id_from_seed_nonce_rpidhash(
    seed: bytes, nonce: bytes, rp_id_hash: bytes, ext_state: bytes = bytes()
) -> bytes:
    assert_32_bytes(seed)
    assert_32_bytes(nonce)
    assert_32_bytes(rp_id_hash)
    assert_bytes(ext_state)
    assert len(ext_state) <= 256

    message = VERSION + nonce + ext_state
    bound_message = rp_id_hash + message
    mac = HMAC(seed, bound_message)
    credential_id = message + mac
    return credential_id


def nonce_extstate_mac_from_credential_id(credential_id: bytes):
    assert_credential_id(credential_id)
    nonce = credential_id[1:33]
    ext_state = credential_id[33:-32]
    mac = credential_id[-32:]

    return nonce, ext_state, mac


def validate_credential_id(seed: bytes, credential_id: bytes, rp_id_hash: bytes):
    assert_32_bytes(seed)
    nonce, ext_state, mac = nonce_extstate_mac_from_credential_id(credential_id)

    message = VERSION + nonce + ext_state
    calculated_mac = HMAC(seed, rp_id_hash + message)

    assert calculated_mac == mac, f"{calculated_mac} != {mac}"
    return True



import pytest

import seedweed


def test_seed_constructor():
    from seedweed import assert_32_bytes as assert_valid_seed

    with pytest.raises(AssertionError):
        assert_valid_seed(bytes([0] * 31))
    with pytest.raises(AssertionError):
        assert_valid_seed(bytes([0] * 33))
    with pytest.raises(AssertionError):
        assert_valid_seed([0] * 31)
    with pytest.raises(AssertionError):
        assert_valid_seed("0" * 32)


def test_zero_seed_zero_nonce():
    zero_seed = bytes([0] * 32)
    zero_mac = bytes([0] * 32)

    scalar, point, keypair, iterations = seedweed.keypair_from_seed_mac(
        zero_seed, zero_mac
    )

    assert iterations == 1, iterations
    assert (
        scalar
        == 19437864054269200754977681106621014646072784646180000699501739860764553555251
    )


def test_multi_iteration():
    seed = bytes.fromhex(
        "dd2ca3b88f9491c042fcc04c5e732e9f6fd9c0eb6f3b99cddd4ae96d661ada2c"
    )
    assert seedweed.H(b"seedweed") == seed
    nonce = bytes.fromhex(
        "8b911917c0b74f77e6fb819d6a8034a94f0fca487cacb41e89235c5c5220947b"
    )
    rp_id = "solokeys.com"
    rp_id_hash = seedweed.H(rp_id.encode())
    mac = seedweed.HMAC(seed, rp_id_hash + bytes([1]) + nonce)
    iterations = seedweed.keypair_from_seed_mac(seed, mac)[-1]

    assert iterations == 2



seed = bytes.fromhex("59ece2f2bcd6512eea7a6bd1694cf1e73552fbf314e219b06f3e2f864f0d99ed")
assert len(seed) == 32
unique_id = bytes.fromhex("4eed5a99990c4cc9d6065556599383495b316262095b7bb65d37c23ec53e7712")
assert len(unique_id) == 32

def HMAC(key, msg):
    import hashlib
    import hmac
    mac = hmac.new(key, digestmod=hashlib.sha256)
    mac.update(msg)
    return mac.digest()

msg = bytes([1]) + unique_id
mac = HMAC(seed, msg)
secret_key = HMAC(seed, mac)

import ecdsa
n = int.from_bytes(secret_key, "little")
assert n >= ecdsa.NIST256p.order
print("found a weirdo!")
print(n)
print(ecdsa.NIST256p.order)
print(n - ecdsa.NIST256p.order)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['seedweed']

package_data = \
{'': ['*']}

install_requires = \
['ecdsa >= 0.16.0']

entry_points = \
{'console_scripts': ['generate-seedweed-test-vectors = '
                     'seedweed.vectors:generate']}

setup(name='seedweed',
      version='1.0rc3',
      description='Seeded WebAuthn Credentials: reference implementation',
      author='Nicolas Stalder',
      author_email='nicolas@solokeys.com',
      url='https://github.com/solokeys/seedweed',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='~=3.6',
     )
